/* -*-C-*-
********************************************************************************
*
* File:         findcrd.c
* RCS:          $Header: $
* Description:  Finds the name of a guitar chord. The input is transformed to
*               offsets (in frets) of the notes wrt. to the lowest string.
*               Next a jump table is created, this contains for each half note,
*               the distance up to the next note found in the input pattern.
*               The jump pattern is concatenated to itself. The patterns in
*               the chord database are matched with the jump table. If the 
*               chord is not matched exactly, a set of alterations is tried
*               to yield a resolution of the chord. 
*               The resolutions output are prefixed with a number indicating
*               the "quality" of the resolution. Meant to be used as a sorting
*               criteria.
* Bugs:         Does only make ONE alteration to each chord in the database.
*               Several others, I am sure ;-)
*               Please report them if you find them (bugs that is.)
* Author:       Staal Vinterbo (staalv@idt.unit.no)
* Created:      Mon Jun  5 13:36:08 1995
* Modified:     Fri Sep 22 17:51:05 2006 (Staal Vinterbo) staal@MGV
* Language:     C
* Package:      N/A
* Status:       EXPERIMENTAL
*
* (c) Copyright 1995-2006, Staal Vinterbo, all rights reserved.
*
* findcrd.c is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* findcrd.c is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with findcrd.c; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*
*
********************************************************************************
*
* Revisions:
*
* Fri Sep 22 14:36:47 2006 (Staal Vinterbo) staal@MGV
*  Added scale/mode resolution. Cleaned up command line code and
*  added options. Added add9 chord to data base. Added GPL Licence.
* Wed Jun 21 17:20:45 1995 (Staal Vinterbo) staal@gartimus.gartnet
*  fixed another oops in main: iterative was not set to 0 if maxdist 
*  was given in  non-interactive mode. Removed unused variable delta
*  in match. Changed default maxdist value to 1 (from 2).
* Sat Jun 10 23:15:14 1995 (Staal Vinterbo) staal@gartimus.gartnet
*  Added 1 to delta if root != bass note. Added incrementing of maxdist
*  if no resolution found and no maxdist value is given. Fixed oops in 
*  main.
* Fri Jun  9 15:26:50 1995 (Staal Vinterbo) staal@gartimus.gartnet
*  Made sure max and min are undef'ed before defining them. Added bass note
*  to named chords resolved to a different root than the lowest tone given
*  in the input pattern. Added version string. Cleaned up some output code.
********************************************************************************
*/

char version[] = "FINDCHORD (findcrd.c) v.2.00, (c) SAV 1995-2006";

#include <stdio.h>
#include <assert.h>  /* assert, define NDEBUG to get rid of this */
#include <string.h>  /* memset, memcpy, strcat */
#include <stdlib.h>  /* atoi, in a pinch: replace with sprintf */

#include <unistd.h>

#ifdef DEBUG
# define DEB(s) s
# define NODEB(s)
# define PI(off,str,s,a,i) {                        \
    int _c;                                         \
    fprintf(str, s);                                \
    for(_c = 0; _c < i; _c++)                       \
	fprintf(str, " %d",(a[_c] + off) % 12);     \
    fprintf(str, "\n");                             \
  }
#else
# define DEB(s)
# define NODEB(s) s
# define PI(off,str,s,a,i)
#endif

/* want to make sure max and min do what I want */
#ifdef min
#undef min
#endif

#define min(a, b) ((a) < (b) ? (a) : (b))

#ifdef max
#undef max
#endif

#define max(a, b) ((a) > (b) ? (a) : (b))

#define STRINGS 6           /* guitar */
#define MAXDIST 11           /* default max number of halfsteps to move note */

/*****************/

struct chord
{
  int len;            /* number of notes actually in this chord */
  int pos[STRINGS];   /* the positions of the notes (normalized) */
  char * name;
};

struct chord chords[] = /* the chord example database */
  {
    {3, {0,4,7,0,0,0},""},
    {3, {0,3,7,0,0,0},"m"},
    {4, {0,4,7,10,0,0},"7"},
    {3, {0,3,6,0,0,0},"dim"},
    {3, {0,4,8,0,0,0},"aug"},
    {4, {0,3,6,8,0,0},"dim6"},
    {4, {0,4,8,9,0,0},"aug6"},
    {4, {0,3,7,11,0,0},"minmaj7"},
    {4, {0,4,7,11,0,0},"maj7"},
    {4, {0,3,7,10,0,0},"m7"},
    {4, {0,3,6,9,0,0},"dim7"},
    {4, {0,3,6,10,0,0},"m7b5"},
    {4, {0,4,7,9,0,0},"6"},
    {4, {0,3,7,9,0,0},"m6"},
    {3, {0,2,7,0,0,0},"sus2"},
    {3, {0,5,7,0,0,0},"sus4"},
    {4, {0,2,7,9,0,0},"6sus2"},
    {4, {0,5,7,9,0,0},"6sus4"},
    {4, {0,2,7,10,0,0},"7sus2"},
    {4, {0,5,7,10,0,0},"7sus4"},
    {4, {0,2,7,11,0,0},"maj7sus2"},
    {4, {0,5,7,11,0,0},"maj7sus4"},
    {4, {0,4,8,11,0,0},"augb7"},          /* standard name? */
    {2, {0,7,0,0,0,0}, "5"},
    {4, {0,2,4,7,0,0}, "add9"},
    {5, {0,2,4,7,10,0},"9"},
    {5, {0,2,3,7,10,0},"m9"},
    {5, {0,2,4,7,11,0},"maj9"},
    {5, {0,2,3,6,9,0},"dim9"},
    {6, {0,2,4,5,7,10},"11"},
    {6, {0,2,3,5,7,10},"m11"},
    {6, {0,2,4,5,7,11},"maj11"},
    {6, {0,2,3,5,6,9},"dim11"},
    {6, {0,2,4,7,9,10},"13"},
    {6, {0,2,3,7,9,10},"m13"},
    {6, {0,2,4,7,9,11},"maj13"},
    {0, {0,0,0,0,0,0}, "noname"}          /* end marker */
  };

#define NSCALES 5
#define MAXNOTES 7         /* number of notes in a scale */
/*****************/

struct scale
{
  int len;
  int pattern[12];    /* the positions of the notes (normalized) */
  char * name;
  char * modenames[MAXNOTES]; /* the names of the modes (if any) */
  int notepos[MAXNOTES];
};

struct scale scales[NSCALES] = 
  {
    { /* Major */
      7, /* len */
      {1,0,1,0,1,1,0,1,0,1,0,1}, /* pattern */
      "Major Scale", /* name */
      {"Ion","Dor","Phryg","Lyd","Mixolyd","Aeol","Locr"}
      /* mode names */,
      {0,2,4,5,7,9,11} /* note positions */
    },
    { /* Ascending (Melodic) Minor */
      7, /* len */
      {1,0,1,1,0,1,0,1,0,1,0,1}, /* pattern */
      "Mel Minor",
      {"min-ma", "dor-b2", "Ly-aug", "Ly-dom", "Hindu", "Loc-#2", "Sup-Loc"}, /* mode names */
      {0,2,3,5,7,9,11} /* note positions */
    },
    { /* Harmonic Minor */
      7, /* len */
      {1,0,1,1,0,1,0,1,1,0,0,1}, /* pattern */
      "Har Minor",
      {"I", "II", "III", "IV", "V", "VI", "VII"}, /* mode names */
      {0,2,3,5,7,8,11} /* note positions */
    },
    { /* Hungarian Minor */
      7, /* len */
      {1,0,1,1,0,0,1,1,1,0,0,1}, /* pattern */
      "Hun Minor",
      {"I", "II", "III", "IV", "V", "VI", "VII"}, /* mode names */
      {0,2,3,6,7,8,11} /* note positions */
    },
    { /* Enigmatic */
      7, /* len */
      {1,1,0,0,1,0,1,0,1,0,1,1}, /* pattern */
      "Enigmatic",
      {"I", "II", "III", "IV", "V", "VI", "VII"}, /* mode names */
      {0,1,4,6,8,10,11} /* note positions */
    }
  };


int inpattern[12];    /* compressed input pattern */ 
int matchpattern[24]; /* scale pattern copied twice */
int indicate_bass = 0;
int othreshold = 2;
int oThreshold = 2;

char * notenames[12] = {"C","C#","D","D#","E","F","F#","G","G#","A","A#","B"}; 
char * sharpnotenames[12] =
  {"1","#1","2","#2","3","4","#4","5","#5","6","#6","7"}; 
char * bnotenames[12] =
  {"1","b2","2","b3","3","4","b5","5","b6","6","b7","7"}; 
char * addnotenames[12] =
  {"1","#1","9","#9","3","11","#11","5","#5","13","#13","7"}; 
char * addbnotenames[12] =
  {"1","b9","9","b3","3","11","b5","5","b13","13","b7","7"}; 

int lowstring = 4;                      /* index in notename array (4 == E) */
int stringdiff[STRINGS] = {5, 5, 5, 4, 5, 0};        /* number of halftones */
/* between the strings */
int jumps[24];      /* foreach halftone, the distance to the next note */
/* in the pattern, duplicated */
int pat_len = 0;    /* the length of the pattern */
int minusones[STRINGS] = {-1, -1, -1, -1, -1, -1}; /* STRINGS times -1 */
int frets[STRINGS] = {-1, -1, -1, -1, -1, -1};     /* the input fret values, */
/* -1 is unused */
int offsets[STRINGS] = {-1, -1, -1, -1, -1, -1};  
int offmap[24];
int maxdist = MAXDIST;   /* the distance cutoff value */
char * pname;
int bass_note = 0;      /* the lowest note */
int matched = 0;        /* count matched patterns */

/* protos */
void prepare_jumps(); 
int match(struct chord *); /* returns non-zero if something was matched */
int main(int, char **);
void usage_exit();
void output(int missing, int nohit,int relroot, struct chord * c, int bass);

/* dirty details below */

void prepare_jumps(char ** args)
{
  int i = 0, j = 0, d = 0, t = 0, used[12];
    
  pat_len = STRINGS;
  memset(used, 0, sizeof(used));
  memset(inpattern, 0, sizeof(inpattern));
  memcpy(frets, minusones, sizeof(frets));
    
  /* calculate fretvalues of all notes on the lowest string (offsets) */
  t = 10000;                      /* a high value */
  for(i = 0; i < STRINGS; i++)
    {
      assert(args + i);
      if(args[i][0] == 'x')
	pat_len--;                /* unused */
      else
	{
	  int m;
	  frets[i] = atoi(args[i]);
	  if(frets[i] + d < t)      /* find the lowest note */
	    {
	      bass_note = j;
	      t = frets[i] + d;
	    }
	  m = (frets[i] + d) % 12;
	  inpattern[(frets[i] + d + lowstring) % 12] = 1;
	  if(!(used[m]++))          /* dont need same note more then once */
	    offsets[j++] = m;
	}
      d += stringdiff[i];           /* next string */
    }

  assert(j <= pat_len);
  pat_len = j;

  /* find the bass note */
  bass_note = (offsets[bass_note] + lowstring) % 12;

  /* sort offsets */
  for(i = 0; i < pat_len; i++)
    {
      d = i;
      for(j = i + 1; j < pat_len; j++)
	if(offsets[j] < offsets[d])
	  d = j;
      if(d > i)
	{
	  t = offsets[i];
	  offsets[i] = offsets[d];
	  offsets[d] = t;
	}
    }

  /* debug stuff */
  PI(0,stderr,"used: ", used, 12);   
  PI(0,stderr,"offsets: ", offsets, pat_len); 
  PI((12-offsets[0]),stderr,"offsets norm: ", offsets, pat_len);

  /* calculate, for each halfnote in the octave, the distance to the */
  /* next note in the input pattern (i.e. input offsets 0 4 7 (a major */
  /* chord) the jumptable becomes 0 3 2 1 0 2 1 0 4 3 2 1. Note that the */
  /* table is contructed such that the first value is a 0.) A 0 value */
  /* denotes that this note is in the input pattern. The offset map, */
  /* offmap, is the mapping from the 0 values to the corresponding */
  /* index in the offset array. (For the jumptable above this is */
  /* 0 - - - 1 - - 2 - - - - , a - may be any number as it is not used) */

  t = 0;
  for(i = 0; i < pat_len - 1; i++)
  {
    jumps[t] = 0;
    offmap[t] = i;
    d = offsets[i+1] - offsets[i];
    for(j = 1; j <= d; j++)
	    jumps[++t] = d - j;
  }

  jumps[t] = 0;
  offmap[t] = i;
    
  for(i = t + 1; i < 12; i ++)
    jumps[i] = 12 - i;

  PI(0,stderr,"jumps: ", jumps, 12);

  /* duplicate to upper 12 so that we can use it to match the chord */
  /* patterns in the chords array with any positive displacement not */
  /* larger than 12 */
  memcpy(jumps + 12, jumps, 12 * sizeof(int));
  memcpy(offmap + 12, offmap, 12 * sizeof(int));
}

int match(struct chord * c)
{
  int mind = 0, /* the minimal displacement that yields at least one match */
    jump = 0,   /* the current total dispacement */
    hits,       /* number of matches found using the current displacement */
    i, j;
  int nohit = -1; /* the index of the note in the chord pattern not */
                  /* matched in the input pattern */
  int root = 0;   /* the fret value of the root note relative to the low */
                  /* string */
  int offshit[STRINGS]; /* used to record if this offset (input pattern) */
                        /* was matched  */
  matched = 0;          /* count matched patterns */

  if((c->len > pat_len + 1) /* chord has more then one tone too much */ ||
     (pat_len - c->len > 1) /* more than one alteration */)
  {
    return matched; /* 0 */
  }

  //  fprintf(stderr, "looking at %s...\n", c->name);

  while(jump + mind < 12) /* need only to match lower 12 of jumps */
  {
    jump += mind;       /* jump ahead to next possible match */
    mind = 13;          /* bigger that possible */
    hits = 0;
    nohit = -1;
    memset(offshit, 0 , sizeof(offshit));

    for(i = 0; i < max(pat_len, c->len); i++) 
    {
	    j = jumps[c->pos[i] + jump];

	    /* j now contains 0 or the number of halftones to the next note */
	    /* in the pattern */
	    
	    if(!j)  /* this was a match */
	    {
        hits++;
        offshit[offmap[c->pos[i] + jump]]++; /* record the hit */
        continue;
	    }
	    /* else */          
	    nohit = i;   /* the index of the note not matched in the chord */
	    if(j < mind) /* find the least jump forward to the next position */
        mind = j;/* that gives at least one match */
    }

    if(hits >= c->len - 1) /* filter out the positions that do not match */
                           /* in all positions but one */
    {
	    int missing = -1; /* the offset of the note in the pattern */
                        /* that is not matched in the chord */
	    root = (jump + offsets[0]) % 12; /*calculate the root */

	    /* find (if it exists) the note in the pattern that is not */
	    /* matched by a note in the chord */
	    for(i = 0; i < pat_len; i++)
        if(!offshit[i])   
        {
          missing = (offsets[i] + 12 - root) % 12;
          assert(missing >= 0);
          break;      /* there is only one of these, so we need */
                      /* not look further */
        }

	    /* nohit contains the index in the c->pos array of the */
	    /* note in the chord not matched in the pattern */
	    /* the resolved chord is named by "unifying" missing and */
	    /* nohit. */

	    output(missing, nohit, root, c, bass_note);

    }
	
  }
  return matched;
}

void output(int missing, int nohit,int relroot, struct chord * c, int bass)
{
  char buf[BUFSIZ], format[BUFSIZ], cc = '#';
  int delta = 0, i, val;

  /*  fprintf(stderr, "missing: %d, nohit: %d, relroot: %s, bass: %s\n",
          missing, nohit, notenames[(relroot + lowstring) % 12],
          notenames[bass]);
  */

  sprintf(buf, "%s%s", notenames[(relroot + lowstring) % 12], c->name);
  if((val = (missing != -1) | ((nohit != -1) << 1)) != 0)
  {
    if(maxdist == 0) /* only exact matches */
	    return;
    i = strlen(buf);
    switch(val)
    {
	  case 1: /* missing != -1  && nohit == -1 => add missing to chrd */
	    assert(missing != -1  && nohit == -1);
	    delta = 1;
	    sprintf(buf + i,"(add%s)",addnotenames[missing]);
	    break;
	  case 2: /* missing == -1  && nohit != -1 remove nohit from chord */
	    assert(missing == -1  && nohit != -1);
	    delta = 1;
	    sprintf(buf + i,"(no %s)",bnotenames[c->pos[nohit]]);
	    break;
	  case 3: /* missing != -1  && nohit != -1 */
	    assert(missing != -1  && nohit != -1);
	    if(pat_len != c->len)
        return; /* we have too few notes in the chord */
	    delta = missing - c->pos[nohit];
	    delta = delta < 0 ? cc = 'b', -delta : delta;
	    assert(delta < 12);
	    if(delta > maxdist)
        return;
	    memset(format, cc, delta);
	    format[delta++] = 0;
	    sprintf(buf + i,"(%s%s)",format, bnotenames[c->pos[nohit]]);
    }
  }
    
  if((relroot + lowstring) % 12 != bass)
  {
    strcat(buf, "/");
    strcat(buf, notenames[bass]);
    delta++;
  }
  if(delta <= othreshold){
    fprintf(stdout, "%d %s", delta, buf);
    NODEB(fprintf(stdout, "\n"));
  }
  matched++;
}

int nsmaller(int * a, int * b, int len)
{
  int dist = 0;
  int i = 0;
  for(i = 0; i < len; i++)
    if(a[i] < b[i])
      dist++;
  return dist;
}

void modes()
{
  int dthreshold = oThreshold;
  int i;

  for(i = 0; i < NSCALES; i++){
    int j = 0;
    int mpattern[24];
    memcpy(mpattern, scales[i].pattern, sizeof(scales[i].pattern));
    memcpy(mpattern + 12, scales[i].pattern, sizeof(scales[i].pattern));
    
    //    fprintf(stderr, "prepared mpattern...\n");
    int phead = 0;
    for(j = 0; j < 12; j++){
      /*      fprintf(stderr, "checking %s, j = %d, mode = %d\n",
              scales[i].name, j, mode); */
      int d = nsmaller(mpattern + j, inpattern, 12);
      if(d > dthreshold)
        continue;
      if(phead == 0){
        printf(" \n \t");
        int k = 0;
        for(k = 0; k < scales[i].len; k++)
          printf("\t%2s", scales[i].modenames[k]);
        printf("\n");
        phead = 1;
      }
      printf("%d %s \t",
             d, scales[i].name);
      int k = 0;
      for(k = 0; k < scales[i].len; k++){
        int nn = (12 - j + scales[i].notepos[k]) % 12;
        printf("%2s", notenames[nn]);
        if(indicate_bass && nn == bass_note)
          printf("*");
        printf("\t");
      }
      printf("\n");
    }
  }
}



int main(int argc, char ** argv) 
{
  char ** cp = argv + 1;
  struct chord * c = chords;
  int i;

  int do_scales = 0;
  int do_both = 0;
  
  pname = argv[0];

  int ochar;
  extern int optind;
  extern char * optarg;

  while((ochar = getopt(argc, argv, "vsSbhm:t:T:")) != EOF){
    switch(ochar){
    case 's':
      do_scales = 1;
      break;
    case 'S':
      do_both = 1;
      break;
    case 'b':
      indicate_bass = 1;
      break;
    case 'm':
      if(sscanf(optarg, "%d", &maxdist) != 1 || maxdist < 0 || maxdist > 12){
        fprintf(stderr,
                "could not read valid distance from -m option %s.", optarg);
        exit(1);
      }
      break;
    case 't':
      if(sscanf(optarg, "%d", &othreshold) != 1 || othreshold < 0
         || othreshold > 12){
        fprintf(stderr,
                "could not read valid threshold from -t option %s.", optarg);
        exit(1);
      }
      break;
    case 'T':
      if(sscanf(optarg, "%d", &oThreshold) != 1 || oThreshold < 0
         || oThreshold > 12){
        fprintf(stderr,
                "could not read valid threshold from -T option %s.", optarg);
        exit(1);
      }
      break;
    case 'v':
	    fprintf(stdout, "%s\n", version);
	    exit(0);
      break;
    case 'h':
    default:
      usage_exit();
    }
  }
  if(argc - optind != 6){
    fprintf(stderr,"incorrect number of arguments given, try -h for help");
    exit(1);
  }

  /* set up patterns */
  cp = &argv[optind];
  prepare_jumps(cp);

  if(do_scales){
    modes();
    exit(0);
  }

  do
  {
    i = 0;
    c = chords;
    while(c->len)
      i += match(c++);
  }
  while(i == 0 && maxdist++ < 12);

  if(!i)
    puts("sorry.");
  else if(do_both)
    modes();
  
  exit(0);
}
    
void usage_exit()
{
  fprintf(stdout, "Usage: %s [ options ] <values>\n"
          "   <values> are the space-separated fret positions for each string,\n"
          "     'x' for unused.\n"
          "   -v prints the version string.\n"
          "   -h prints a short usage summary.\n"
          "   -t <val> sets the chords output value threshold to <val>.\n"
          "   -T <val> sets the scales output value threshold to <val>.\n"
          "   -s performs scale resolution.\n"
          "   -S performs scale resolution in addition to chord resolution.\n"
          "   -b puts a star next to mode matching bass lowest note.\n"
          "   -m <maxdistance> sets the maximal distance you want the altered\n"
          "     note to move. %d is default. Addition or subtraction of a note\n"
          "     counts as 1.\n"
          "   Each resolution is prefixed by a number that is\n"
          "   derived from this distance. Lower value means smaller\n"
          "   alteration to the chord from which the resolution was found.\n",
          pname,maxdist);
  exit(1); 
}


/* that's all folks */
